part of tecfy_core_client_package;

class ConnectionService {
  static List<Connection> _connections;
  static StreamController<List<Connection>> _connectionsStreamController;
  static StreamController<Connection> _activeConnectionStreamController;
  static Stream<Connection> _activeConnectionStream;
  static Stream<List<Connection>> _connectionsStream;
  static Connection _activeConnection;
  static bool _isBusy = false;
  static bool _isInit = false;

  static Stream<Connection> get activeConnectionStream {
    if (_activeConnectionStream == null)
      _activeConnectionStream =
          _activeConnectionStreamController.stream.asBroadcastStream();
    return _activeConnectionStream;
  }

  static Stream<List<Connection>> get connectionsStream {
    if (_connectionsStream == null)
      _connectionsStream =
          _connectionsStreamController.stream.asBroadcastStream();
    return _connectionsStream;
  }

  static Connection get activeConnection => _activeConnection;
  static bool get isOnline => activeConnection != null;
  static bool get isOffline => activeConnection == null;
  static bool get isOfflineMode =>
      DatabaseService.isSupportLocalDB &&
      LocalConfigService.operationMode != OperationMode.online;
  static bool get isOnlineMode =>
      !DatabaseService.isSupportLocalDB ||
      LocalConfigService.operationMode == OperationMode.online;

  static Future<void> init(List<Connection> c) async {
    _connectionsStreamController = StreamController<List<Connection>>();
    _activeConnectionStreamController = StreamController<Connection>();

    var result = await DatabaseService.getEntities('_ServerConnections');
    if (result != null && result.length > 0) {
      _connections =
          result.map<Connection>((con) => Connection.fromJson(con)).toList();
      _connections.sort((a, b) => a.sortId - b.sortId);
    } else {
      _connections = c;
      var saveConnsResult = await DatabaseService.saveEntity(
          '_ServerConnections',
          _connections.map((con) => con.toJson()).toList(),
          overwriteLocal: true);
    }
    _connections.sort((a, b) => a.sortId - b.sortId);
    _connectionsStreamController.add(_connections);
    _checkActiveConnection();
    Timer.periodic(Duration(seconds: 10), checkAllConnections);
    checkAllConnections(null);
  }

  static reloadConnections() async {
    var result = await DatabaseService.getEntities('_ServerConnections');
    if (result != null && result.length > 0) {
      _connections =
          result.map<Connection>((con) => Connection.fromJson(con)).toList();
      _connections.sort((a, b) => a.sortId - b.sortId);
      _connectionsStreamController.add(_connections);
      checkAllConnections(null);
    }
  }

  static _checkActiveConnection() {
    var ac = _connections.firstWhere(
        (c) => c.active && c.status == ConnectionStatus.connected,
        orElse: () => null);

    if (ac?.apiUrl != _activeConnection?.apiUrl) {
      bool appWasOfffline = ConnectionService.isOffline;
      print(
          '###################  active connection Changed to : ${ac?.name} #########################');

      _activeConnection = ac;
      _activeConnectionStreamController.add(ac);
      TecfyCoreApp.apiURL = ac.apiUrl;
      TecfyCoreApp.wsURL = 'ws://${ac.server}/ws';
      if (appWasOfffline && ConnectionService.isOnline) {
        TecfySyncService.syncNow();
      }
    }

    if (activeConnection != null) _isInit = true;
  }

  static void checkAllConnections(Timer timer) async {
    try {
      if (_isBusy) return;
      _isBusy = true;
      var changed = false;
      //print('================== start check connection');
      for (var con in _connections) {
        if (!con.active) continue;

        //print('================== start check con ${con.name}');
        OperationResult<String> result;
        try {
          result = await ApiService.httpGetString('${con.apiUrl}Version')
              .timeout(Duration(seconds: con.isCloud ? 5 : 2));
        } catch (err) {
          result = OperationResult<String>(
              success: false,
              message: err.toString().contains('Timeout')
                  ? 'Connection Timeout'
                  : err.roString());
        }
        //print('================== end check con ${con.name}');
        if (result.success) {
          if (con.status != ConnectionStatus.connected ||
              con.version != result.data.replaceAll('"', '')) {
            changed = true;
          }
          con.status = ConnectionStatus.connected;
          con.version = result.data.replaceAll('"', '');
          con.lastActive = DateTime.now().toUtc();
        } else {
          //print('================== con not working ${con.name}');
          if (con.status != ConnectionStatus.connected) {
            changed = true;
          }
          con.status = ConnectionStatus.disconnected;
          con.errorMessage = result.message ?? '';
          if (con.errorMessage.isEmpty) {
            con.errorMessage = 'Unknown Error! ${con.apiUrl}Version';
          }
        }
        _checkActiveConnection();
      }
      if (changed) {
        DatabaseService.saveEntity(
            '_ServerConnections', _connections.map((e) => e.toJson()).toList(),
            overwriteLocal: true);
      }
      if (activeConnection == null && !_isInit) {
        print(
            'XXXXXXXXXXXXXXXXXXXXXXXXXXXXX  Not active connection XXXXXXXXXXXXXXXXXXXXXXXXXXXXXX');
        _isInit = true;
        TecfyCoreApp.apiURL = 'http://null/';
        TecfyCoreApp.wsURL = 'ws://null/ws';
        //AppNavigator.navigateTo(context, creator)
      }
      _checkActiveConnection();
    } catch (err) {
      print('XXXXXXXX checkAllConnections Error: $err');
    }
    _isBusy = false;
  }

  static Widget getConnectionIconStream() {
    return StreamBuilder(
        stream: activeConnectionStream,
        builder: (c, snap) => getConnectionIcon(c));
  }

  static Widget getSettingWithConnectionIconStream() {
    return StreamBuilder(
        stream: activeConnectionStream,
        builder: (c, snap) => getSettingConnectionIcon(c));
  }

  static Widget getConnectionIcon(BuildContext context) {
    Widget icon;
    if (_activeConnection == null)
      icon = CircleAvatar(
          child: Icon(Icons.close, color: Colors.white, size: 20),
          backgroundColor: Colors.red,
          radius: 20);
    else
      icon = _getIconWidget(_activeConnection);

    return GestureDetector(
        child: icon, onTap: () => _showStatusDialog(context));
  }

  static Widget getSettingConnectionIcon(BuildContext context) {
    Widget icon;
    if (_activeConnection == null)
      icon = CircleAvatar(
          child: Icon(Icons.close, color: Colors.white, size: 20),
          backgroundColor: Colors.red,
          radius: 20);
    else
      icon = _getIconWidget(_activeConnection);

    return GestureDetector(
        child: icon, onTap: () => _showStatusDialog(context));
  }

  static Widget _getIconWidget(Connection con) {
    return CircleAvatar(
        child: Icon(con.isCloud ? Icons.cloud_done : Icons.compare_arrows,
            color: Colors.white, size: 20),
        backgroundColor: con.color,
        radius: 20);
  }

  static void _showStatusDialog(BuildContext context) async {
    await AppNavigator.showAppDialog(
      context,
      (c) => Center(
        child: Material(
          child: Container(
            height: 400,
            width: 300,
            child: Column(children: [
              Expanded(
                  child: ListView(
                      children: _connections
                          .map<Widget>((con) => ListTile(
                                title: Text('${con.name} ${con.version ?? ''}',
                                    style: TextStyle(
                                        color: con.status ==
                                                ConnectionStatus.connected
                                            ? Colors.black
                                            : Colors.red)),
                                subtitle: Text(con.errorMessage ?? '',
                                    style: TextStyle(fontSize: 10)),
                                trailing: _getIconWidget(con),
                              ))
                          .toList())),
              TextButton(
                  child: Text(AppLocale.translate(c, 'Settings')),
                  onPressed: () async {
                    if (_isBusy) return;
                    _isBusy = true;
                    await AppNavigator.navigateTo(
                        c, () => EntityListPage('_ServerConnections'));

                    await reloadConnections();
                    AppNavigator.pop(context);
                    _isBusy = false;
                  }),
            ]),
          ),
        ),
      ),
    );
  }
}
